/*
:vim:ts=4:
 *
 * For the FUTABA 7 segment VFD on the Artisan Media centre case
 *
 * Attempt at a kernal module to talk to the Futaba VFD. This is for my
 * mythdora 5.0 system
 *
 * This is very much a work in progress, please help out if you can
 *
 * This driver when it has grabbed the LCD will print its device
 * major number in the system log. You need to create a char device at that node
 * and then any string written to that device will appear on the display
 *
 *
 */

//---------------------INCLUDES------------------------------------------------
#include <linux/init.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/kdev_t.h>
#include <linux/usb.h>
#include <linux/cdev.h>
#include <linux/hid.h>
#include <linux/fs.h>
#include <asm/uaccess.h>

//---------------------DEFINES-------------------------------------------------
#define DISP
#ifdef DISP
	#define VENDOR_ID 0x0547
	#define PRODUCT_ID 0x7000

// A mouse used for inital tests

#else
	#define VENDOR_ID 0x15ca
	#define PRODUCT_ID 0x00c3
#endif

#define DISPLAY_SIZE 10
#define DISPLAY_STRING_OPCODE 0x8b


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Steve Williams");

//--------------------Function declarations------------------------------------
static int futaba_init(void);
static int futaba_probe(struct usb_interface *interface, const struct usb_device_id *id);
static void futaba_exit(void);
static void futaba_disconnect(struct usb_interface *interface);
ssize_t futaba_write(struct file *filp, char __user *buff, size_t count, loff_t *offp);
int futaba_send_string(char *string);
int futaba_send_report(void);

/*
 * This is a botch. The kernel does not export "usbhid_modify_dquirk()" which we
 * need because at start up the hid driver grabs the display and crashes it by asking
 * it for a report.
 *
 * Before running this module you need to locate your Symbols.map file (usually in
 * /boot) and fill in the value below to match *your* kernel.
 *
 * You also need to make sure that the hid driver has been detached from the LCD
 * after loading this module. Then unlod this module and re-load it.
 *
 * I know it is messy - this is a beginning I am working on a way to reset the device
 * from this driver but can't see how to do it yet.
 *
 * If your usbhid driver is a module then life for you is easy just "rmmod" it
 */
static int (*p_usbhid_modify_dquirk)(uint16_t vendor,uint16_t product,uint32_t quirk) = 
0xc05b6746;

//---------------------Globals-------------------------------------------------
struct VFDContext_t
{
	int initialised;
	dev_t deviceNum;
	struct usb_device *dev;
	int pipe;
	uint8_t *report;
}VFDContext;

// Device ID table for our VFD
static struct usb_device_id vfd_table[]=
{
	{ USB_DEVICE(VENDOR_ID,PRODUCT_ID) },
	{}
};

MODULE_DEVICE_TABLE(usb,vfd_table);

static struct usb_driver vfd_driver=
{
	.name = "futaba",
	.id_table = vfd_table,
	.probe = futaba_probe,
	.disconnect = futaba_disconnect,
};

static struct file_operations futaba_fops = 
{
	.owner = THIS_MODULE,
	.write = &futaba_write,
};

//------------------Implimentation---------------------------------------------
int futaba_send_report(void)
{
	int n;

	n = usb_control_msg(	VFDContext.dev,
				usb_sndctrlpipe(VFDContext.dev,0),
				HID_REQ_SET_REPORT,	// request
				0x21,			// Request type SET_REPORT
				0x0200,			// Report type ( TYPE | ID 0)
				0,			// windex
				VFDContext.report,	// data
				64,			// size
				5000);			// timeout

	printk(KERN_ALERT "usb_control_msg() said %d\n",n);

	return(n);
}

int futaba_send_string(char *string)
{
	int n;
	int len;

	len = strlen(string);

	if(len > DISPLAY_SIZE)
	{
		len = DISPLAY_SIZE;
	}

	memset(VFDContext.report,0,64);

//	Format the report message to send a string
	VFDContext.report[0] = DISPLAY_STRING_OPCODE;
	VFDContext.report[1] = 0x04;
	VFDContext.report[2] = 0x01;
	VFDContext.report[3] = DISPLAY_SIZE;

//	Copy in the string
	for(n=0; n < len; n++)
	{
		VFDContext.report[n+4] = string[n];
	}

	n = futaba_send_report();

	return(n != 64);
}

static int futaba_init(void)
{
	int retVal = -1;
	int n;
	struct cdev *futaba_cdev;

	printk(KERN_ALERT "Futaba VFD driver\n");

//	Register with the USB subsystem
	if( (retVal = usb_register(&vfd_driver)) )
	{
		err("usb_register failed. Error = %d",retVal);
	}

	alloc_chrdev_region(&VFDContext.deviceNum,0,1,"futaba");

	printk(KERN_ALERT "We go dev %d Major\n",MAJOR(VFDContext.deviceNum));

	futaba_cdev = cdev_alloc();
	futaba_cdev->ops = &futaba_fops;

	cdev_add(futaba_cdev,VFDContext.deviceNum,1);

	n = p_usbhid_modify_dquirk(VENDOR_ID,PRODUCT_ID,HID_QUIRK_NOGET ) ;
	printk(KERN_ALERT "usbhid_modify_dquirk() said %d\n",n);

	return retVal;
}

static int futaba_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct usb_device *dev = NULL;
	struct usb_host_interface *iface_desc = NULL;
	int num_endpoints;
	int retVal = -1;

	printk(KERN_ALERT "futaba_probe() looks like we have our device\n");

	dev = usb_get_dev(interface_to_usbdev(interface));
	iface_desc = interface->cur_altsetting;
	num_endpoints = iface_desc->desc.bNumEndpoints;

	printk(KERN_ALERT "We got interface %d\n",iface_desc->desc.bInterfaceNumber);

	if( iface_desc->desc.bInterfaceNumber == 0 )
	{
		printk(KERN_ALERT "Binding to this one\n");
		VFDContext.pipe = usb_sndctrlpipe(dev,0);
		VFDContext.dev = dev;
		VFDContext.report = kmalloc(64,GFP_KERNEL);
		VFDContext.initialised++;
//		usb_reset_device(VFDContext.dev);
		retVal = 0;
	}

	return retVal;
}


ssize_t futaba_write(struct file *filp, char __user *buff, size_t count, loff_t *offp)
{
	char string[128];

	if( VFDContext.initialised == 0 )
	{
		printk(KERN_ALERT "Not doing anything - not yet initialiased\n");
		return count;
	}

	memset(string,' ',sizeof(string));

	if( copy_from_user(string,buff,count) )
	{
		return -EFAULT;
	}

	string[DISPLAY_SIZE] = 0;

	futaba_send_string(string);

	return count;
}

static void futaba_disconnect(struct usb_interface *interface)
{
	printk(KERN_ALERT "Disconnect function\n");

}

static void futaba_exit(void)
{
	printk(KERN_ALERT "Futaba driver exit\n");

//	Un register
	unregister_chrdev_region(VFDContext.deviceNum,1);

	kfree(VFDContext.report);
	usb_deregister(&vfd_driver);
}

module_init(futaba_init);
module_exit(futaba_exit);
